package com.mobiledefense.verify.validate.ui.presenter

import com.mobiledefense.common.util.Maybe
import com.mobiledefense.common.util.Observer
import com.mobiledefense.common.util.ParallelAsyncTask
import com.mobiledefense.lean.EventsTracker
import com.mobiledefense.lean.data.model.Analytic
import com.mobiledefense.registration.validator.RegistrationValidator
import com.mobiledefense.verify.features.VerificationFeatureManager
import com.mobiledefense.verify.request.api.VerificationRequestApi
import com.mobiledefense.verify.update.UpdatePhoneNumber
import com.mobiledefense.verify.validate.api.VerificationValidationApi
import com.mobiledefense.verify.validate.helper.VerificationPinValidator
import com.mobiledefense.verify.validate.helper.VerificationValidationNavigationHelper
import com.mobiledefense.verify.validate.ui.model.VerificationValidationViewModel
import com.mobiledefense.verify.validate.ui.view.VerificationValidationView.Action
import java.lang.ref.WeakReference

class VerificationValidationViewPresenterImpl @JvmOverloads constructor(
    private val featureManager: VerificationFeatureManager,
    private val pinValidator: VerificationPinValidator,
    private val validateApi: VerificationValidationApi,
    private val requestApi: VerificationRequestApi,
    private val navigationHelper: VerificationValidationNavigationHelper,
    private val eventsTracker: EventsTracker,
    private var currentState: VerificationValidationViewModel,
    observer: Observer<VerificationValidationViewModel>? = null,
    private val updatePhoneNumber: UpdatePhoneNumber? = null)
    : VerificationValidationViewPresenter {

    companion object {
        const val SHOULD_VERIFY_NUMBER = true

        private const val DEFAULT_COUNTRY_CODE = "US"

        private fun createInitialState(lineNumber: String, id: String): VerificationValidationViewModel {
            return VerificationValidationViewModel.InputError(lineNumber, id, "")
        }
    }

    override var viewModelObserver: Observer<VerificationValidationViewModel>? = observer
        set(observer) {
            if (observer == null) {
                onMoveToForegroundTask?.cancel(true)
                onMoveToForegroundTask = null

                validateTask?.cancel(true)
                validateTask = null

                requestTask?.cancel(true)
                requestTask = null
            } else {
                observer.next(currentState)

                if (onMoveToForegroundTask == null) {
                    onMoveToForegroundTask = OnMoveToForegroundTask(this).apply { executeInParallel() }
                }
            }
            field = observer
        }
    override val pinChangesObserver: Observer<Action.PinChange> by lazy { PinChangesObserver() }
    override val validateButtonTapsObserver: Observer<Action.ValidateButtonTap> by lazy { ValidateButtonTapsObserver() }
    override val requestButtonTapsObserver: Observer<Action.RequestButtonTap> by lazy { RequestButtonTapsObserver() }
    override val requestErrorDialogDismissalsObserver: Observer<Action.RequestErrorDialogDismissal> by lazy { RequestErrorDialogDismissalsObserver() }
    override val backTapsObserver: Observer<Action.BackTap> by lazy { BackTapsObserver() }
    override val smsMessageObserver: Observer<String> by lazy { SmsMessageObserver() }
    private var onMoveToForegroundTask: OnMoveToForegroundTask? = null
    private var validateTask: ValidateTask? = null
    private var requestTask: RequestTask? = null

    @JvmOverloads constructor(
        featureManager: VerificationFeatureManager,
        pinValidator: VerificationPinValidator,
        validateApi: VerificationValidationApi,
        requestApi: VerificationRequestApi,
        navigationHelper: VerificationValidationNavigationHelper,
        eventsTracker: EventsTracker,
        lineNumber: String,
        id: String,
        updatePhoneNumber: UpdatePhoneNumber? = null)
            : this(featureManager, pinValidator, validateApi, requestApi, navigationHelper,
                   eventsTracker, createInitialState(lineNumber, id), null, updatePhoneNumber)

    private class OnMoveToForegroundTask(presenter: VerificationValidationViewPresenterImpl) : ParallelAsyncTask<Void, Void, Boolean>() {

        private val presenterRef = WeakReference(presenter)

        override fun doInBackground(vararg ignore: Void): Boolean {
            return presenterRef.get()?.featureManager?.isVerificationEnabled ?: false
        }

        override fun onPostExecute(isFeatureEnabled: Boolean) {
            if (!isFeatureEnabled) {
                presenterRef.get()?.navigationHelper?.close()
            }

            presenterRef.get()?.onMoveToForegroundTask = null
        }
    }

    private inner class PinChangesObserver : Observer<Action.PinChange>() {

        override fun onResult(action: Action.PinChange) {
            val state = currentState.takeIf {
                it is VerificationValidationViewModel.Input
                || it is VerificationValidationViewModel.InputError
                || it is VerificationValidationViewModel.ValidationFailure
            } ?: return

            currentState = when (pinValidator.validate(action.pin)) {
                RegistrationValidator.Status.VALID -> VerificationValidationViewModel.Input(
                    state.lineNumber, state.id, action.pin)
                else -> VerificationValidationViewModel.InputError(state.lineNumber, state.id, action.pin)
            }
            viewModelObserver?.next(currentState)
        }
    }

    private inner class ValidateButtonTapsObserver : Observer<Action.ValidateButtonTap>() {

        override fun onResult(ignore: Action.ValidateButtonTap) {
            val state = currentState.takeIf {
                it is VerificationValidationViewModel.Input
                || it is VerificationValidationViewModel.ValidationFailure
            } ?: return

            eventsTracker.track(Analytic.Builder()
                .withEvent(Analytic.Event.VERIFICATION_CODE_CONTINUE_TAPPED)
                .build())

            if (validateTask == null) {
                validateTask = ValidateTask(state.lineNumber, state.id, state.pin,
                    this@VerificationValidationViewPresenterImpl).apply { executeInParallel() }
            }
        }
    }

    private class ValidateTask(private val lineNumber: String,
                               private val id: String,
                               private val pin: String,
                               presenter: VerificationValidationViewPresenterImpl)
                               : ParallelAsyncTask<Void, Void, ValidateTaskState>() {

        private val presenterRef = WeakReference(presenter)

        override fun onPreExecute() {
            presenterRef.get()?.reconcileValidateTaskState(when {
                isCancelled -> ValidateTaskState.Canceled
                else -> ValidateTaskState.Start
            })
        }

        override fun doInBackground(vararg ignore: Void): ValidateTaskState {
            if (!isCancelled) {
                presenterRef.get()?.let { presenter ->
                    val result = presenter.validateApi.validateVerificationPin(id, pin)

                    if (!isCancelled) {
                        return if (result.isLeft) {
                            ValidateTaskState.Error(result.left.value)
                        } else {
                            val updateResult = presenter.updatePhoneNumber?.update(
                                lineNumber, Maybe.just(result.right),
                                SHOULD_VERIFY_NUMBER
                            )?.hasValue()
                            when (updateResult) {
                                false -> ValidateTaskState.Error("Verification failed")
                                else -> ValidateTaskState.Success(result.right)
                            }
                        }
                    }
                }
            }
            return ValidateTaskState.Canceled
        }

        override fun onPostExecute(taskState: ValidateTaskState) {
            presenterRef.get()?.reconcileValidateTaskState(when {
                isCancelled -> ValidateTaskState.Canceled
                else -> taskState
            })

            presenterRef.get()?.validateTask = null
        }
    }

    private sealed class ValidateTaskState {

        object Start : ValidateTaskState()
        data class Success(val id: String) : ValidateTaskState()
        data class Error(val message: String? = null) : ValidateTaskState()
        object Canceled : ValidateTaskState()
    }

    private fun reconcileValidateTaskState(taskState: ValidateTaskState) {
        var newState: VerificationValidationViewModel? = null

        if ((currentState is VerificationValidationViewModel.Input
                    || currentState is VerificationValidationViewModel.ValidationFailure)
                    && taskState === ValidateTaskState.Start) {
            newState = VerificationValidationViewModel.Validating(currentState.lineNumber,
                currentState.id, currentState.pin)
        } else if (currentState is VerificationValidationViewModel.Validating) {
            if (taskState is ValidateTaskState.Success) {
                newState = VerificationValidationViewModel.ValidationSuccess(currentState.lineNumber,
                    currentState.id, currentState.pin)
                navigationHelper.close(currentState.lineNumber, taskState.id)
            } else if (taskState is ValidateTaskState.Error) {
                newState = VerificationValidationViewModel.ValidationFailure(currentState.lineNumber,
                    currentState.id, currentState.pin,
                    VerificationValidationViewModel.ValidationFailure.Reason(taskState.message))
            } else if (taskState === ValidateTaskState.Canceled) {
                newState = VerificationValidationViewModel.Input(currentState.lineNumber,
                    currentState.id, currentState.pin)
            }
        }

        newState?.let {
            currentState = it
            viewModelObserver?.next(it)
        }
    }

    private inner class RequestButtonTapsObserver : Observer<Action.RequestButtonTap>() {

        override fun onResult(ignore: Action.RequestButtonTap) {
            val state = currentState.takeIf {
                it is VerificationValidationViewModel.Input
                || it is VerificationValidationViewModel.InputError
                || it is VerificationValidationViewModel.ValidationFailure
            } ?: return

            eventsTracker.track(Analytic.Builder()
                .withEvent(Analytic.Event.RESEND_CODE_TAPPED)
                .build())

            if (requestTask == null) {
                requestTask = RequestTask(state.lineNumber,
                    this@VerificationValidationViewPresenterImpl).apply { executeInParallel() }
            }
        }
    }

    private class RequestTask(private val lineNumber: String,
                              presenter: VerificationValidationViewPresenterImpl)
                              : ParallelAsyncTask<Void, Void, RequestTaskState>() {

        private val presenterRef = WeakReference(presenter)

        override fun onPreExecute() {
            presenterRef.get()?.reconcileRequestTaskState(when {
                isCancelled -> RequestTaskState.Canceled
                else -> RequestTaskState.Start
            })
        }

        override fun doInBackground(vararg ignore: Void): RequestTaskState {
            if (!isCancelled) {
                presenterRef.get()?.let { presenter ->
                    val result = presenter.requestApi.requestVerificationPin(DEFAULT_COUNTRY_CODE, lineNumber)
                    if (!isCancelled) {
                        return when {
                            result.isLeft -> RequestTaskState.Error(result.left.value)
                            else -> RequestTaskState.Success(result.right)
                        }
                    }
                }
            }
            return RequestTaskState.Canceled
        }

        override fun onPostExecute(taskState: RequestTaskState) {
            presenterRef.get()?.reconcileRequestTaskState(when {
                isCancelled -> RequestTaskState.Canceled
                else -> taskState
            })

            presenterRef.get()?.requestTask = null
        }
    }

    private sealed class RequestTaskState {

        object Start : RequestTaskState()
        data class Success(val id: String) : RequestTaskState()
        data class Error(val message: String? = null) : RequestTaskState()
        object Canceled : RequestTaskState()
    }

    private fun reconcileRequestTaskState(taskState: RequestTaskState) {
        var newState: VerificationValidationViewModel? = null

        if ((currentState is VerificationValidationViewModel.Input
                    || currentState is VerificationValidationViewModel.InputError)
                    && taskState === RequestTaskState.Start) {
            newState = VerificationValidationViewModel.Requesting(currentState.lineNumber,
                currentState.id, currentState.pin)
        } else if (currentState is VerificationValidationViewModel.ValidationFailure
            && taskState === RequestTaskState.Start) {
            newState = VerificationValidationViewModel.Requesting(currentState.lineNumber, currentState.id,
                currentState.pin, (currentState as VerificationValidationViewModel.ValidationFailure).reason)
        } else if (currentState is VerificationValidationViewModel.Requesting) {
            val validationFailureReason = (currentState as VerificationValidationViewModel.Requesting).validationFailureReason

            if (taskState is RequestTaskState.Error) {
                newState = VerificationValidationViewModel.RequestError(
                    currentState.lineNumber, currentState.id, currentState.pin,
                    taskState.message, validationFailureReason)
            } else {
                if (validationFailureReason != null) {
                    if (taskState is RequestTaskState.Success) {
                        newState = VerificationValidationViewModel.ValidationFailure(
                            currentState.lineNumber, taskState.id, currentState.pin,
                            validationFailureReason)
                    } else if (taskState === RequestTaskState.Canceled) {
                        newState = VerificationValidationViewModel.ValidationFailure(
                            currentState.lineNumber, currentState.id, currentState.pin,
                            validationFailureReason)
                    }
                } else {
                    if (taskState is RequestTaskState.Success) {
                        newState = when (pinValidator.validate(currentState.pin)) {
                            RegistrationValidator.Status.VALID -> VerificationValidationViewModel.Input(
                                currentState.lineNumber, taskState.id, currentState.pin)
                            else -> VerificationValidationViewModel.InputError(currentState.lineNumber,
                                taskState.id, currentState.pin)
                        }
                    } else if (taskState === RequestTaskState.Canceled) {
                        newState = when (pinValidator.validate(currentState.pin)) {
                            RegistrationValidator.Status.VALID -> VerificationValidationViewModel.Input(
                                currentState.lineNumber, currentState.id, currentState.pin)
                            else -> VerificationValidationViewModel.InputError(currentState.lineNumber,
                                currentState.id, currentState.pin)
                        }
                    }
                }
            }
        }

        newState?.let {
            currentState = it
            viewModelObserver?.next(it)
        }
    }

    private inner class RequestErrorDialogDismissalsObserver : Observer<Action.RequestErrorDialogDismissal>() {

        override fun onResult(ignore: Action.RequestErrorDialogDismissal) {
            val state = currentState as? VerificationValidationViewModel.RequestError ?: return

            currentState = if (state.validationFailureReason != null) {
                VerificationValidationViewModel.ValidationFailure(state.lineNumber,
                    state.id, state.pin, state.validationFailureReason)
            } else {
                when (pinValidator.validate(state.pin)) {
                    RegistrationValidator.Status.VALID -> VerificationValidationViewModel.Input(
                        state.lineNumber, state.id, state.pin)
                    else -> VerificationValidationViewModel.InputError(state.lineNumber, state.id, state.pin)
                }
            }
            viewModelObserver?.next(currentState)
        }
    }

    private inner class BackTapsObserver : Observer<Action.BackTap>() {

        override fun onResult(ignore: Action.BackTap) {
            navigationHelper.close()
        }
    }

    private inner class SmsMessageObserver : Observer<String>() {

        override fun onResult(message: String) {
            val state = currentState.takeIf {
                it is VerificationValidationViewModel.Input
                || it is VerificationValidationViewModel.InputError
                || it is VerificationValidationViewModel.ValidationFailure
            } ?: return

            val matcher = VerificationPinValidator.VALID_PATTERN.matcher(message)

            val pin = when {
                matcher.find() && matcher.groupCount() == 1 -> matcher.group(1)
                else -> return
            }

            currentState = VerificationValidationViewModel.Input(state.lineNumber, state.id, pin)
            viewModelObserver?.next(currentState)

            if (validateTask == null) {
                validateTask = ValidateTask(state.lineNumber, state.id, pin,
                    this@VerificationValidationViewPresenterImpl).apply { executeInParallel() }
            }
        }
    }
}
