package com.mobiledefense.verify.validate.ui.view

import android.content.Context
import android.support.v7.app.AlertDialog
import android.text.Editable
import android.text.TextWatcher
import android.view.View
import android.view.ViewGroup
import android.widget.Button
import android.widget.FrameLayout
import android.widget.TextView
import com.dev.mobiledefense.R
import com.mobiledefense.base.ui.control.StatusEditText
import com.mobiledefense.base.util.PhoneNumberUtilsImpl
import com.mobiledefense.common.util.Observer
import com.mobiledefense.home.ui.control.DashedSpinner
import com.mobiledefense.verify.validate.ui.model.VerificationValidationViewModel
import com.mobiledefense.verify.validate.ui.view.VerificationValidationView.Action

class VerificationValidationViewImpl(context: Context): FrameLayout(context), VerificationValidationView {

    override val viewModelObserver: Observer<VerificationValidationViewModel> by lazy { ViewModelObserver() }
    override var pinChangesObserver: Observer<Action.PinChange>? = null
    override var validateButtonTapsObserver: Observer<Action.ValidateButtonTap>? = null
    override var requestButtonTapsObserver: Observer<Action.RequestButtonTap>? = null
    override var requestErrorDialogDismissalsObserver: Observer<Action.RequestErrorDialogDismissal>? = null
    private val messageTextView: TextView
    private val pinInputView: StatusEditText
    private val validateButton: Button
    private val requestButton: Button
    private val validatingLayout: ViewGroup
    private val validatingLoadingIndicator: DashedSpinner
    private val validatingStatusTextView: TextView
    private var requestErrorDialog: AlertDialog? = null
    private var isInternalPinChange: Boolean = false

    init {
        View.inflate(context, R.layout.verification_validation, this)

        messageTextView = findViewById(R.id.verification_validation_message)

        pinInputView = findViewById(R.id.verification_validation_pin_input)
        pinInputView.addTextChangedListener(object : TextWatcher {

            override fun onTextChanged(text: CharSequence, start: Int, before: Int, count: Int) {
                if (!isInternalPinChange) {
                    pinChangesObserver?.next(Action.PinChange(text.toString()))
                }

                isInternalPinChange = false
            }

            override fun afterTextChanged(editable: Editable) {}
            override fun beforeTextChanged(text: CharSequence, start: Int, count: Int, after: Int) {}
        })

        validateButton = findViewById(R.id.verification_validation_validate_button)
        validateButton.setOnClickListener {
            validateButtonTapsObserver?.next(Action.ValidateButtonTap)
        }

        requestButton = findViewById(R.id.verification_validation_request_button)
        requestButton.setOnClickListener {
            requestButtonTapsObserver?.next(Action.RequestButtonTap)
        }

        validatingLayout = findViewById(R.id.verification_validation_validating_layout)
        validatingLoadingIndicator = validatingLayout.findViewById(R.id.verification_validation_validating_loading_indicator)
        validatingStatusTextView = validatingLayout.findViewById(R.id.verification_validation_validating_status)
        validatingLayout.visibility = View.GONE
    }

    private inner class ViewModelObserver : Observer<VerificationValidationViewModel>() {

        override fun onResult(viewModel: VerificationValidationViewModel) {
            messageTextView.text = resources.getString(R.string.verification_validation_message)
                .replace("{phone_number}", PhoneNumberUtilsImpl().getFormattedNumber(viewModel.lineNumber))

            if (pinInputView.text != viewModel.pin) {
                isInternalPinChange = true
                pinInputView.text = viewModel.pin
            }

            when (viewModel) {
                is VerificationValidationViewModel.Input -> {
                    pinInputView.isEnabled = true
                    pinInputView.setStatusToDefault()

                    validateButton.isEnabled = true
                    requestButton.isEnabled = true

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.stopSpin()
                    validatingLayout.visibility = View.GONE
                }
                is VerificationValidationViewModel.InputError -> {
                    pinInputView.isEnabled = true
                    pinInputView.setStatusToDefault()

                    validateButton.isEnabled = false
                    requestButton.isEnabled = true

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.stopSpin()
                    validatingLayout.visibility = View.GONE
                }
                is VerificationValidationViewModel.Requesting -> {
                    pinInputView.isEnabled = false

                    val validationFailureReason = viewModel.validationFailureReason
                    if (validationFailureReason != null) {
                        pinInputView.setStatusToError(validationFailureReason.message
                                ?: resources.getString(R.string.verification_server_error_default_message))
                    } else {
                        pinInputView.setStatusToDefault()
                    }

                    validateButton.isEnabled = false
                    requestButton.isEnabled = false

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.stopSpin()
                    validatingLayout.visibility = View.GONE
                }
                is VerificationValidationViewModel.RequestError -> {
                    pinInputView.isEnabled = false

                    val validationFailureReason = viewModel.validationFailureReason
                    if (validationFailureReason != null) {
                        pinInputView.setStatusToError(validationFailureReason.message
                                ?: resources.getString(R.string.verification_server_error_default_message))
                    } else {
                        pinInputView.setStatusToDefault()
                    }

                    validateButton.isEnabled = false
                    requestButton.isEnabled = false

                    showRequestErrorDialog(viewModel.message)

                    validatingLoadingIndicator.stopSpin()
                    validatingLayout.visibility = View.GONE
                }
                is VerificationValidationViewModel.Validating -> {
                    pinInputView.isEnabled = false
                    pinInputView.setStatusToDefault()

                    validateButton.isEnabled = false
                    requestButton.isEnabled = false

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.fastSpin()
                    validatingLoadingIndicator.setInnerImage(R.drawable.error_view_loading)
                    validatingStatusTextView.text = resources.getString(R.string.verification_validation_status_in_progress)
                    validatingLayout.visibility = View.VISIBLE
                }
                is VerificationValidationViewModel.ValidationSuccess -> {
                    pinInputView.isEnabled = false
                    pinInputView.setStatusToDefault()

                    validateButton.isEnabled = false
                    requestButton.isEnabled = false

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.stopSpin()
                    validatingLoadingIndicator.setInnerImage(R.drawable.error_view_thumbs_up)
                    validatingStatusTextView.text = resources.getString(R.string.verification_validation_status_success)
                    validatingLayout.visibility = View.VISIBLE
                }
                is VerificationValidationViewModel.ValidationFailure -> {
                    pinInputView.isEnabled = true

                    pinInputView.setStatusToError(viewModel.reason.message
                            ?: resources.getString(R.string.verification_server_error_default_message))

                    validateButton.isEnabled = true
                    requestButton.isEnabled = true

                    dismissRequestErrorDialog()

                    validatingLoadingIndicator.stopSpin()
                    validatingLayout.visibility = View.GONE
                }
            }
        }
    }

    private fun showRequestErrorDialog(message: String?) {
        if (requestErrorDialog == null) {
            requestErrorDialog = AlertDialog.Builder(context)
                .setTitle(R.string.verification_server_error_title)
                .setMessage(message ?: resources.getString(R.string.verification_server_error_default_message))
                .setOnCancelListener {
                    requestErrorDialogDismissalsObserver?.next(Action.RequestErrorDialogDismissal)
                }
                .show()
        }
    }

    private fun dismissRequestErrorDialog() {
        requestErrorDialog?.dismiss()
        requestErrorDialog = null
    }
}
